# This module is a user interface to the device driver acl8112
# * Copyright by Juergen Prenzel (jprenze@gwdg.de) 2001
# * Version 0.1
# * This software is published under the GPL (see file LICENSE)
# **************************************************************/

package acl8112;

use FileHandle;
require "ioctl.ph";

sub new{
    my $class = shift;

    my $self = {};
    bless $self,$class;

    $self->{DEVS} = {};

    return $self;

}

sub open{
    my $self = shift;
    my $dev = shift;
    my $arg = "";
    my ($rc);

    $fh = new FileHandle($dev,"r+");
    if(! $fh){
	warn "Cannot open $dev\n";
	return $fh;
    }

    $rc = ioctl($fh,&IOCTL_CTYP,$arg) || -1;
    warn "$!\n" if $rc != 0;
    my $typ = unpack "a",$arg;
    if($typ ne "H" and $typ ne "D" and $typ ne "P"){
	warn "Unknown card type \"$typ\" for $dev.\n";
	return undef;
    }
    $self->{DEVS}{$fh}{TYP} = $typ;

    $rc = ioctl($fh,&IOCTL_AIR,$arg) || -1;
    warn "$!\n" if $rc != 0;
    my $air = unpack "s",$arg;
    if($air != 5 and $air != 10){
	warn "Unknown analog input range \"$air\" for $dev.\n";
	return undef;
    }
    $self->{DEVS}{$fh}{AIR} = $air;

    $rc = ioctl($fh,&IOCTL_AOR,$arg) || -1;
    warn "$!\n" if $rc != 0;
    my $aor = unpack "s",$arg;
    $self->{DEVS}{$fh}{AOR} = $aor;

    $rc = ioctl($fh,&IOCTL_BP,$arg) || -1;
    warn "$!\n" if $rc != 0;
    my $bp = unpack "s",$arg;
    $self->{DEVS}{$fh}{BP} = $bp;

    my @stat = stat($fh);
    my $minor = $stat[6];
    my $opr = $minor & 0x3f;
    $self->{DEVS}{$fh}{OPR} = $opr;

    $self->{DEVS}{$fh}{DEV} = $dev;

    $self->gain($fh,1) or die;

    return $fh;
}

sub get{
    my $self = shift;
    my $fh = shift;
    my($rc,$val,$v,$min,$gain,$aor,$bp,$volt,$raw);

    if($self->{DEVS}{$fh}{OPR} < 16){  # analog input 
	$rc = sysread($fh,$val,2);
	if($rc != 2){
	    warn "read error in ",$self->{DEVS}{$fh}{DEV},"\n";
	    return undef;
	} 
	$v = unpack "S",$val;

	$gain = $self->{DEVS}{$fh}{GAIN};

	if($self->{DEVS}{$fh}{TYP} eq "P"){
	    $aor = $self->{DEVS}{$fh}{AOR};
	    $bp = 1;
	}
	else{
	    $aor = 10;
	    $bp = $self->{DEVS}{$fh}{BP};
	}

	if($bp){
	    $min = -$aor/2/$gain;
	    $raw = $v -2048;
	}
	else{
	    $min = 0;
	    $raw = $v;
	}
	$volt = $min + $v*$aor/$gain/4096;
	return ($volt,$raw);
    }
    elsif($self->{DEVS}{$fh}{OPR} == 18){ #digital input
	$rc = sysread($fh,$val,2);
	if($rc != 2){
	    warn "read error in ",$self->{DEVS}{$fh}{DEV},"\n";
	    return undef;
	} 
	$v = unpack "S",$val;
	return $v;
    }
    else{
	warn "$self->{DEVS}{$fh}{DEV} is a read-only-device!\n";
	return undef;
    }

}

sub put{
    my $self = shift;
    my $fh = shift;
    my $val = shift;

    my($rc,$v,$opr,$arg);

    $opr = $self->{DEVS}{$fh}{OPR};
    if($opr != 16 and $opr != 17 and $opr != 19){
	warn "$self->{DEVS}{$fh}{DEV} is a write-only-device!\n";
	return undef;
    }
    elsif($opr == 19){   # digital output
	$arg = pack "S", $val;
	$rc = syswrite($fh,$arg);
	if($rc != 2){
	    warn "write error in ",$self->{DEVS}{$fh}{DEV},"\n";
	    return undef;
	} 
	return 2;
    }
    else{  # analog output
	my $hi = $self->{DEVS}{$fh}{AOR};
	if($val > $hi or $val <0){
	    warn "Analog output range 0 ... $hi Volt!\n";
	    return undef;
	}
	$v = int($val/$hi*4096);
	$arg = pack "S", $v;
	$rc = syswrite($fh,$arg);
	if($rc != 2){
	    warn "write error in ",$self->{DEVS}{$fh}{DEV},"\n";
	    return undef;
	} 
	return 2;
    }
}

sub gain{              # set the gain factor for analog input
    my $self = shift;
    my $fh = shift;
    my $gain = shift;
    my ($v);

    my $bp = $self->{DEVS}{$fh}{BP};
    my $typ = $self->{DEVS}{$fh}{TYP};

    if($typ eq "H"){
	if($bp){
	    if($gain == 1)       {$v = 0;}
	    elsif($gain == 10)   {$v = 1;}
	    elsif($gain == 100)  {$v = 2;}
	    elsif($gain == 1000) {$v = 3;}
	    elsif($gain == 0.5)  {$v = 8;}
	    elsif($gain == 5)    {$v = 9;}
	    elsif($gain == 50)   {$v = 10;}
	    elsif($gain == 500)  {$v = 11;}
	    else{
		warn "Gain $gain not defined for HG & biploar\n";
		return undef;
	    }
	}
	else{
	    if($gain == 1)       {$v = 4;}
	    elsif($gain == 10)   {$v = 5;}
	    elsif($gain == 100)  {$v = 6;}
	    elsif($gain == 1000) {$v = 7;}
	    else{
		warn "Gain $gain not defined for HG & uniploar\n";
		return undef;
	    }
	}
    }
    elsif($typ eq "D"){
	if($bp){
	    if($gain == 0.5)    {$v = 8;}
	    elsif($gain == 1)   {$v = 0;}
	    elsif($gain == 2)   {$v = 1;}
	    elsif($gain == 4)   {$v = 2;}
	    elsif($gain == 8)   {$v = 3;}
	    else{
		warn "Gain $gain not defined for DG & biploar\n";
		return undef;
	    }
	}
	else{
	    if($gain == 1)      {$v = 4;}
	    elsif($gain == 2)   {$v = 5;}
	    elsif($gain == 4)   {$v = 6;}
	    elsif($gain == 8)   {$v = 7;}
	    else{
		warn "Gain $gain not defined for DG & unipolar\n";
		return undef;
	    }
	}
    }
    elsif($typ eq "P"){
	if($gain == 1)      {$v = 0;}
	elsif($gain == 2)   {$v = 1;}
	elsif($gain == 4)   {$v = 2;}
	elsif($gain == 8)   {$v = 3;}
	elsif($gain == 16)   {$v = 4;}
	else{
	    warn "Gain $gain not defined for PG\n";
	    return undef;
	    }
    }

    $self->{DEVS}{$fh}{GAIN} = $gain;
    $arg = pack "s",$v; 
    $rc = ioctl($fh,&IOCTL_GAIN,$arg);
    if(! $rc){
	warn "$! (rc=$rc)";
	return undef;
    }
    return 1;
}



#----------------------------------------------------------------------

=pod

=head1 NAME

acl8112 - interface to the ACL-8112 cards

=head1 SYNOPSIS

 use lib "/usr/local/acl8112";      # or wherever the module resides
 use acl8112; 
 $acl = new acl8112;

 $c1 = $acl->open("/dev/aclAai0");
 $acl->gain($c1,10);                # set gain
 ($volt,$raw) = $acl->get($c1);     # read from A/D

 $c2 = $acl->open("/dev/aclBao0");
 $acl->put($c2,0.333);              # set D/A to 0.333 Volt

 $c3 = $acl->open("/dev/aclAdo");
 $acl->put($c3,1);                  # set bit 0 in TTL output

=head1 DESCRIPTION

The perl module acl8112 is a companion to the kernel module
acl8112. The latter is a insmod loadable device driver for the
ACL-8112 card. These cards do 12 bit A/D conversion on 16 channels (or
8 channels bipolar), 12 bit D/A conversion on 2 channels and digital
(TTL) input and output on 16 channels. 

The kernel module must have been loaded, possibly by the perl script
insmod.pl, in order for the perl module to work. This must, of course,
be done with root privileges.

=head1 METHODS

=over 4  

=item B<open:>

    $chn = $acl->open("/dev/aclAdo")

Opens one of the special files that belong to the device
driver. Returns a file handle.

=item B<get:>

    ($volt, $raw) = $acl->get($chn)
    $bits = $acl->get($chn)

Read a value from one of the 8 or 16 analog input channels or from the
digital input. In case of analog input it returns a 2-element list
with (1st) the volts calculated from (2nd) the raw reading. The raw
reading is an integer in the range of 0 ... 4095 for unipolar and
-2048 ... 2047 for bipolar measuring mode. The raw values can be used
to decide whether the value is on the limits of the measuring range and whether it is adviseable to change the gain.

If the digital input is
read, only one value is returned. It is to be interpreted as a bit-array that corresponds to the 16 TTL channels.

=item B<gain:>

    $acl->($chn, $gain)

Set the gain
for an analog input channel. $gain can be one of the values allowed
for that card type (HG, DG, or PG) and operation mode (bipolar or
unipolar). See card documentation, tables 4.2-1 to 4.2-4.


=item B<put:>

    $acl->put($chn, $volt)
    $acl->put($chn, $bits)

Write to analog or digital output. In case of analog output $volt is
interpreted as volts. In case of digital output $bits is interpreted
as a bit-array that corresponds to the 16 TTL channels.


=back

=head1 AUTHOR

Jrgen Prenzel (jprenze@gwdg.de)


=head1 SEE ALSO

=over 4

=item * the manual to the ACL-8112 series of cards (included in doc/)

=item * the README that comes with the device driver

=back

=head1 DIAGNOSTCS

In case of an error all methods return undef and hopefully write a message on STDERR.


=cut
#####################################################
1;
