/* A driver for the ACL-8112 (HG/DG/PG) cards 
 * Copyright by Juergen Prenzel (jprenze@gwdg.de) 2001
 * Version 0.1
 * This software is published under the GPL (see file LICENSE)
 **************************************************************/
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/fs.h>
#include <linux/malloc.h>
#include <asm/uaccess.h> 
#include <linux/ioport.h> 
#include <linux/ioctl.h> 
#include <linux/errno.h> 
#include <asm/io.h>         /* inb, outb */ 
#include <linux/delay.h> 
#include "acl8112.h"

EXPORT_NO_SYMBOLS;
/* ------------------------------function prototypes-----------*/
int PFX(open) (struct inode *inode, struct file *filp);
int PFX(release) (struct inode *inode, struct file *filp);
ssize_t PFX(airead) (struct file *, char *, size_t, loff_t *);
ssize_t PFX(aowrite) (struct file *filp, const char *buf, size_t count,
		     loff_t *f_pos);
ssize_t PFX(dowrite) (struct file *filp, const char *buf, size_t count,
		      loff_t *f_pos);
ssize_t PFX(diread) (struct file *, char *, size_t, loff_t *);
int PFX(ioctl) (struct inode *inode, struct file *filp,
                 unsigned int cmd, unsigned long arg);

/* ------------------------------ module variables ------------*/
static int Major;      /* major number    */
static int Ncards=1;   /* number of ACL-8112 cards */
static aclgnpar Gain[MAXCARDS][16]; /* gain for each ai channel */

static struct file_operations Aifops = { /* analog input */
  read: PFX(airead),
  ioctl: PFX(ioctl),
  open: PFX(open),
  release: PFX(release)
};
static struct file_operations Aofops = { /* analog output */
  write: PFX(aowrite),
  ioctl: PFX(ioctl),
  open: PFX(open),
  release: PFX(release)
};
static struct file_operations Difops = { /* digital input */
  read: PFX(diread),
  ioctl: PFX(ioctl),
  open: PFX(open),
  release: PFX(release)
};
static struct file_operations Dofops = { /* digital output */
  write: PFX(dowrite),
  ioctl: PFX(ioctl),
  open: PFX(open),
  release: PFX(release)
};


/* ------------------------------ insmod parameters -----------*/
static short io[MAXCARDS] = {0x220,0,0,0};    /* io ports        */
MODULE_PARM(io, "0-4h");
static short bp[MAXCARDS]= {0,0,0,0};         /* bipolar or not  */
MODULE_PARM(bp, "0-4h");
/* card type (HG: 'H'=72, DG: 'D'=68, PG: 'P'=80) */
static char ctp[MAXCARDS] = {'H','H','H','H'}; 
MODULE_PARM(ctp, "0-4b");
static short air[MAXCARDS] = {10,10,10,10}; /* A/D input range (PG only) */
MODULE_PARM(air, "0-4h");
static short aor[MAXCARDS] = {10,10,10,10}; /* D/A output range  */
MODULE_PARM(aor, "0-4h");
static int dbglvl=0;                        /* debug level     */
MODULE_PARM(dbglvl, "0-1i");


/*
 * Open and close
 */

int PFX(open) (struct inode *inode, struct file *filp)
{
  acldev * dev;
  int crd,opr,bas;

  dev = kmalloc(sizeof(acldev), GFP_KERNEL);
  if(!dev) return -ENOMEM;
  filp->private_data = dev;
  
  crd = CARD(inode->i_rdev);
  dev->card = crd;
  opr = OPER(inode->i_rdev);
  dev->oper = opr;
  bas = io[crd];

  PDEBUG(2, "open: card = %i, operation = %i\n",crd,opr);

  if(crd >= MAXCARDS || opr >= MAXOPRS) return -ENODEV;

  /* default gain factor 1.0, see table 4.2-2 in card manual */
  if (bp[crd]) Gain[crd][opr].gain = 0;
  else Gain[crd][opr].gain = 4; 

  /* switch file operations */
  for(;;){
    if(opr < 16){
      outb_p(1,bas+11); /* set to softw. trigger & polling */
      filp->f_op = &Aifops;
      break;
    }
    if(opr < 18){
      filp->f_op = &Aofops;
      break;
    }
    if(opr == 18){
      filp->f_op = &Difops;
      break;
    }
    if(opr == 19){
      filp->f_op = &Dofops;
      break;
    }
    return -ENODEV;
  }
  MOD_INC_USE_COUNT;
  return 0;          /* success */
}

int PFX(release) (struct inode *inode, struct file *filp)
{
  PDEBUG(2, "close\n");
  MOD_DEC_USE_COUNT;
  kfree(filp->private_data);
  return (0);
}

/* read and write */

ssize_t PFX(aowrite) (struct file *filp, const char *buf, size_t count,
		     loff_t *f_pos)
{
  acldev *dev;
  int offset;

  unsigned int b0,b1;
  b0 = buf[0];
  b1 = buf[1];
  PDEBUG(4, "aowrite 0x%x 0x%x\n",b0,b1);

  dev = filp->private_data;
  if(dev->oper == 16) offset = 4;
  else offset = 6;

  outb_p(buf[0], io[dev->card] + offset);
  outb_p(buf[1], io[dev->card] + offset+1);

  return 2;
}

ssize_t PFX(airead) (struct file *filp, char *buf, size_t count,
		     loff_t *f_pos)
{
  acldev *dev;
  int crd,opr,bas,i;
  char val,hi,lo;
  static struct file *oldfilp;

  dev = filp->private_data;
  opr = dev->oper;
  crd = dev->card;
  bas = io[crd];

  PDEBUG(4, "airead (%i %i 0x%x)\n",crd,opr,bas);

  if(oldfilp != filp){
    oldfilp = filp;
    /* set the multiplexer */
    if(bp[crd] == 0){
      if(opr < 8){
	val = 0x10 | opr;
      }
      else{
	val = 0x20 | opr;
      }
    }
    else{
      if(opr < 8){
	val = 0x30 | opr;
      }
      else{
	return -ENODEV;
      }
    }
    outb_p(val,bas+10); /* setting the channel */
    udelay(50);         /* needs some time */

    /* set the gain */
    outb_p(Gain[crd][opr].gain,bas+9);
    udelay(50);          /* also needs some time */
    Gain[crd][opr].newgain = 0;
  }
  if(Gain[crd][opr].newgain){
    outb_p(Gain[crd][opr].gain,bas+9);
    udelay(50);     
    Gain[crd][opr].newgain = 0;
  }
  PDEBUG(4,"gain = %i\n",Gain[crd][opr].gain);

  /* trigger A/D */
  outb_p(0,bas+12);

  /* read A/D */
  for(i=0;i<123456;i++){
    hi = inb_p(bas+5);
    if((hi & 0x10) == 0){
      goto success;
    }
  }
 /* failure: */
  PDEBUG(4, "no response after %i polls.\n",i);
  return 0;
 success:
  lo = inb_p(bas+4);
  PDEBUG(4, "polls: %i, HB = 0x%x, LB = 0x%x\n",i,
	 hi&0xff, lo&0xff);
  buf[1] = hi;
  buf[0] = lo;
  return 2;
}
ssize_t PFX(dowrite) (struct file *filp, const char *buf, size_t count,
		     loff_t *f_pos)
{
  acldev *dev;
  int bas;
  unsigned int b0,b1;
  b0 = buf[0];
  b1 = buf[1];
  PDEBUG(4, "dowrite 0x%x 0x%x\n",b1, b0);

  dev = filp->private_data;
  bas = io[dev->card];
  outb_p(buf[0], bas+13);
  outb_p(buf[1], bas+14);

  return 2;
}
ssize_t PFX(diread) (struct file *filp, char *buf, size_t count, loff_t *fpos)
{
  acldev *dev;
  int bas,hi,lo;

  dev = filp->private_data;
  bas = io[dev->card];
  hi = inb_p(bas+7);
  lo = inb_p(bas+6);
  PDEBUG(4, "diread: HB = 0x%x, LB = 0x%x\n", hi&0xff, lo&0xff);
  buf[1] = hi;
  buf[0] = lo;
  return 2;
}
/*--------------------------------- ioctl ----------------------------*/
int PFX(ioctl) (struct inode *inode, struct file *filp,
                 unsigned int cmd, unsigned long arg)
{
  acldev *dev;
  int crd,opr;
  short gn;

  dev = filp->private_data;
  opr = dev->oper;
  crd = dev->card;

  switch(cmd){
  case IOCTL_GAIN:   /* set gain */
    gn = * (short *) arg;
    PDEBUG(3,"ioctl: gain = %i\n",gn);
    if(bp[crd]){
      if(opr > 7) return -EINVAL;
      if(gn != 0 && gn != 1 && gn != 2 && gn != 3 &&
	 gn != 8 && gn != 9 && gn != 11) return -EINVAL;
    }
    else{
      if(opr > 15) return -EINVAL;
      if(gn != 4 && gn != 5 && gn != 6 && gn != 7) return -EINVAL;
    }
    Gain[crd][opr].gain = gn;
    Gain[crd][opr].newgain = 1;
    break;
  case IOCTL_CTYP:   /* get card type */
    PDEBUG(3,"ioctl: card = %i, ctyp = %c\n",crd,ctp[crd]);
    *(char *)arg = ctp[crd];
    break;
  case IOCTL_BP:   /* get bipolar switch */
    PDEBUG(3,"ioctl: card = %i, bp = %i\n",crd,bp[crd]);
    *(short *)arg = bp[crd];
    break;
  case IOCTL_AIR:   /* get analog input range */
    PDEBUG(3,"ioctl: card = %i, air = %c\n",crd,air[crd]);
    *(short *)arg = air[crd];
    break;
  case IOCTL_AOR:   /* get analog output range */
    PDEBUG(3,"ioctl: card = %i, aor = %c\n",crd,aor[crd]);
    *(short *)arg = aor[crd];
    break;
  default:
    return -EINVAL;
  }


  return 0;
}

/*---------------------------------- module management ----------------*/
int init_module(void)
{
  int i,rc;

  PDEBUG(1, "init module ----------- insmod parameters:\n");

  /* test insmod parameters */
  for(i=0;i<MAXCARDS;i++){
    if(i>0 && io[i-1]==io[i]){
      Ncards = i;
      break;
    }
    if(i>0 && io[i]==0){
      Ncards = i;
      break;
    }
    if(io[i] < 0x200 || io[i] > 0x3ff || io[i] % 0x10 != 0){
      printk(KERN_ERR "ACL8112: Illegal port: 0x%x\n",io[i]);
      return -EINVAL;
    }
    if(bp[i] < 0 || bp[i] > 1){
      printk(KERN_ERR "ACL8112: Illegal bipolar switch: %i\n",bp[i]);
      return -EINVAL;
    }
    if(ctp[i] != 'H' && ctp[i] != 'D' && ctp[i] != 'P' ){
      printk(KERN_ERR "ACL8112: Illegal card type: %c\n",ctp[i]);
      return -EINVAL;
    }
    if(air[i] != 5 && air[i] != 10){
      printk(KERN_ERR "ACL8112: Illegal an. inp. range: %i\n",air[i]);
      return -EINVAL;
    }
  }
  PDEBUG(1, "ports = %x %x %x %x\n",io[0],io[1],io[1],io[1]);
  PDEBUG(1, "bipolar = %x %x %x %x\n",bp[0],bp[1],bp[1],bp[1]);
  PDEBUG(1, "card types = %cG %cG %cG %cG\n",ctp[0],ctp[1],ctp[1],ctp[1]);
  PDEBUG(1, "an.inp.rg. = %i %i %i %i\n",air[0],air[1],air[1],air[1]);
  PDEBUG(1, "an.out.rg. = %i %i %i %i\n",aor[0],aor[1],aor[1],aor[1]);
  PDEBUG(1, "debug level = %i\n",dbglvl);
  PDEBUG(1, "-----------------------\n");

  /* We use dynamic allocation of major number
   */
  Major = register_chrdev(0,QMODNAM, &Aifops);
  if(Major < 0) return Major;

  /* claim io ports */
  for(i=0;i<Ncards;i++){
    if((rc = check_region(io[i],16)) < 0) return rc;
    request_region(io[i],16,QMODNAM);
  }

  return 0;
}

void cleanup_module(void)
{   
  int i; 
  /* release io ports */
  for(i=0;i<Ncards;i++){
    release_region(io[i],16);
  }
  unregister_chrdev(Major,QMODNAM);
  PDEBUG(1, "cleanup end\n");

}
